"""
Created on 17 oct. 2018

@author: Axel Huynh-Phuc, Thibaud Gasser
"""
import traci
from traci._trafficlight import Logic
from typing import Iterable

from shapely.geometry.linestring import LineString

from model import Area, Vehicle


def compute_edge_weight(edge_id):
    
    co2 = traci.edge.getCO2Emission(edge_id)
    co = traci.edge.getCOEmission(edge_id)
    nox = traci.edge.getNOxEmission(edge_id)
    hc = traci.edge.getHCEmission(edge_id)
    pmx = traci.edge.getPMxEmission(edge_id)
    
    return (co2 + co + nox + hc + pmx)
            
def adjust_edges_weights(area): 
    area.weight_adjusted = True    
    for lane in area._lanes:
        edge_id = traci.lane.getEdgeID(lane.lane_id)
        weight = compute_edge_weight(edge_id)  # by default edges weight = length/mean speed
        traci.edge.setEffort(edge_id, weight)
        
    for veh_id in traci.vehicle.getIDList():
        traci.vehicle.rerouteEffort(veh_id)
        
def limit_speed_into_area(area: Area, vehicles: Iterable[Vehicle], speed_rf):
    area.limited_speed = True
    for lane in area._lanes:
        traci.lane.setMaxSpeed(lane.lane_id, speed_rf * lane.initial_max_speed)
    
def modifyLogic(logic, rf): #rf for "reduction factor" 
    new_phases = [] 
    for phase in logic._phases:
        new_phase = traci.trafficlight.Phase(phase.duration*rf,phase.minDuration*rf,phase.maxDuration*rf,phase.phaseDef)
        new_phases.append(new_phase)
        
    return traci.trafficlight.Logic("new-program", 0 , 0 , 0 , new_phases)    

def adjust_traffic_light_phase_duration(area, reduction_factor):
    area.tls_adjusted = True
    for tl in area._tls:
        for logic in tl._logics:
            traci.trafficlights.setCompleteRedYellowGreenDefinition(tl.tl_id, modifyLogic(logic,reduction_factor))
    
def count_vehicles_in_area(area):
    vehicles_in_area = 0 
    for lane in area._lanes:
        vehicles_in_area += traci.lane.getLastStepVehicleNumber(lane.lane_id)
    return vehicles_in_area

def lock_area(area):
    area.locked = True
    for lane in area._lanes:
        traci.lane.setDisallowed(lane.lane_id, 'passenger')
        
def reverse_actions(area):
    #Reset max speed to original 
    if area.limited_speed:
        area.limited_speed = False
        for lane in area._lanes:
            traci.lane.setMaxSpeed(lane.lane_id, lane.initial_max_speed)
    
    #Reset traffic lights initial duration  
    if area.tls_adjusted:
        area.tls_adjusted = False
        for tl in area._tls:
            for initial_logic in tl._logics:
                traci.trafficlights.setCompleteRedYellowGreenDefinition(tl.tl_id, initial_logic._logic)
    
    #Unlock the area 
    if area.locked:
        area.locked = False
        for lane in area._lanes:
            traci.lane.setAllowed(lane.lane_id, '') #empty means all classes are allowed 
            
    